﻿using System.Collections.Generic;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.ProgramSynthesis {
  /// <summary>
  /// Replaces, in the top type vector, item at index(from integer stack) with the first lit-type item.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorSetExpression<T> : StatelessExpression {
    protected VectorSetExpression() { }
    [StorableConstructor]
    protected VectorSetExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IInternalPushInterpreter interpreter, IPushStack<IReadOnlyList<T>> vectorStack, IPushStack<T> literalStack, bool isLiteralTypeInteger = false) {
      return vectorStack.IsEmpty ||
             vectorStack.Top.Count == 0 ||
             literalStack.IsEmpty ||
             (isLiteralTypeInteger && interpreter.IntegerStack.Count < 2) ||
             (!isLiteralTypeInteger && interpreter.IntegerStack.IsEmpty);
    }

    protected void Eval(
      IInternalPushInterpreter interpreter,
      IPushStack<IReadOnlyList<T>> vectorStack,
      IPushStack<T> literalStack,
      bool isLiteralTypeInteger = false) {
      T literal;
      long x;

      if (isLiteralTypeInteger) {
        literal = literalStack.Top;
        x = interpreter.IntegerStack[1];
        interpreter.IntegerStack.Remove(2);
      } else {
        literal = literalStack.Pop();
        x = interpreter.IntegerStack.Pop();
      }

      var vector = vectorStack.Top;
      var newTop = new List<T>(vector);
      var index = x.AsInt(vector.Count);

      if (index < 0)
        index *= -1;

      newTop[index] = literal;
      vectorStack.Top = newTop;
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.IntegerVector,
    "INTEGER[].SET",
    "Replaces in the top INTEGER[] item at index (from INTEGER stack) with the first INTEGER item.",
    StackTypes.Integer)]
  public class IntegerVectorSetExpression : VectorSetExpression<long> {
    public IntegerVectorSetExpression() { }
    [StorableConstructor]
    protected IntegerVectorSetExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.IntegerVectorStack, interpreter.IntegerStack, true);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.IntegerVectorStack, interpreter.IntegerStack, true);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.FloatVector,
    "FLOAT[].SET",
    "Replaces in the top FLOAT[] item at index (from INTEGER stack) with the first FLOAT item.",
    StackTypes.Float | StackTypes.Integer)]
  public class FloatVectorSetExpression : VectorSetExpression<double> {
    public FloatVectorSetExpression() { }
    [StorableConstructor]
    protected FloatVectorSetExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.FloatVectorStack, interpreter.FloatStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.BooleanVector,
    "BOOLEAN[].SET",
    "Replaces in the top BOOLEAN[] item at index (from INTEGER stack) with the first BOOLEAN item.",
    StackTypes.Boolean | StackTypes.Integer)]
  public class BooleanVectorSetExpression : VectorSetExpression<bool> {
    public BooleanVectorSetExpression() { }
    [StorableConstructor]
    protected BooleanVectorSetExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.StringVector,
    "STRING[].SET",
    "Replaces in the top STRING[] item at index (from INTEGER stack) with the first STRING item.",
    StackTypes.String | StackTypes.Integer)]
  public class StringVectorSetExpression : VectorSetExpression<string> {
    public StringVectorSetExpression() { }
    [StorableConstructor]
    protected StringVectorSetExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.StringVectorStack, interpreter.StringStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
